import sys
from argparse import HelpFormatter
from functools import lru_cache, partial
from io import BytesIO
from subprocess import check_output, SubprocessError


def error(message):
    """Write an error to the console"""
    print(f'\033[31mError:\033[0m {message}', file=sys.stderr)
    return 1


def b2s(binary):
    """
    Binary to string helper which ignores all data which can't be decoded
    :param binary: Binary bytes string
    :return: String
    """
    return binary.decode(encoding='ascii', errors='ignore')


def b2s_file(binary_file):
    """
    File based binary to string helper
    :param binary_file: Binary file to convert
    :return: The same file object with a patched method
    """
    orig = binary_file.read
    binary_file.read = lambda *a, **k: b2s(orig(*a, **k))
    return binary_file


def log(message):
    print(message, file=sys.stderr)
    return 0


@lru_cache()
def wc(file):
    """Get number of lines in a file"""
    with open(file, mode='rb') as fd:
        return sum(1 for _ in fd)


def handle_interrupt(func):
    """Decorator which ensures that keyboard interrupts are handled properly."""
    def wrapper():
        try:
            return func() or 0
        except KeyboardInterrupt:
            print('\n\033[31mError:\033[0m Aborted.')
            return 1
    return wrapper


def print_diagnostic_info():
    """
    Get info about the current system used in bug reports
    Please fill out the damn bug reports people, how hard is it
    """
    from stegcracker import __version__

    def run(cmd):
        try:
            return f'$ {cmd}\n' + check_output(cmd, shell=True).decode()
        except SubprocessError:
            return 'unknown'

    err = partial(print, file=sys.stderr)

    err('### Command Used')
    err('```')
    err('$ ' + (' '.join(sys.argv)))
    err('```\n')

    err('### StegCracker Version')
    err('```')
    err(__version__)
    err('```\n')

    err('### StegCracker Type')
    err('```')
    err(run('file $(which stegcracker)'))
    err('```\n')

    err('### StegHide Version')
    err('```')
    err(run('steghide --version'))
    err('```\n')

    err('### Python Version')
    err('```')
    err(sys.version)
    err('```\n')

    err('### System Version')
    err('```')
    err(run('uname -a'))
    err(run('cat /etc/issue'))
    err('```\n')


class DevNull(BytesIO):
    def write(self, *_): pass

    def read(self, *_): return b''


class CustomHelpFormatter(HelpFormatter):
    def __init__(self, prog):
        super().__init__(prog, indent_increment=2, max_help_position=7, width=None)

    # noinspection PyProtectedMember
    def _format_action(self, action):
        result = super(CustomHelpFormatter, self)._format_action(action) + "\n"

        if 'show this help message and exit' in result:
            result = result.replace('show', 'Show', 1)

        return result



import shutil
import sys
import time
from itertools import islice
from subprocess import Popen, PIPE
from traceback import print_tb
from typing import Iterable
from threading import BoundedSemaphore
from multiprocessing.pool import ThreadPool

from stegcracker import __url__
from stegcracker.helpers import error, b2s, b2s_file, print_diagnostic_info


class Cracker:
    SUPPORTED_FILES = ('jpg', 'jpeg', 'bmp', 'wav', 'au')

    def __init__(self, file: str, output: str, line_count: int,
                 threads: int = 8, chunk_size: int = 64, quiet: bool = False,
                 verbose: bool = False):
        """
        Cracker constructor
        :param threads: Number of threads to attempt to crack the signature
        :param file: File to (attempt) to crack
        :param output: Output file to write the file to
        :param chunk_size: Number of passwords to attempt per thread
        """
        self.lock = BoundedSemaphore()
        self.pool = ThreadPool(processes=threads)
        self.thread_count = threads

        self.quiet = quiet
        self.verbose = verbose
        self.file = file
        self.output = output
        self.chunk_size = chunk_size
        self.line_count = line_count or 1
        self.has_error = False
        self.iterable = None

        self.attempts = 0
        self.password = None

    def run(self, iterable: Iterable[str]):
        """Run the brute-forcer by iterating over a set of strings"""
        self.iterable = iterable

        for i in range(self.thread_count):
            self.pool.apply_async(self.crack, args=[i + 1], error_callback=self.error_handler)

        self.pool.close()
        self.pool.join()

    def error_handler(self, exc):
        """
        Error callback handler for thread related issues
        :param exc: Exception
        :return: None
        """
        error(
            f'Unhandled exception in cracker thread. Please report this issue '
            f'on the official bug tracker: "{__url__}/issues" and don\'t forget '
            f'to include the following traceback and diagnostic info:\n')

        print('## Stack Trace', file=sys.stderr)
        print('```', file=sys.stderr)
        print(type(exc).__name__ + ': ' + str(exc), file=sys.stderr)
        print_tb(exc.__traceback__, file=sys.stderr)
        print('```\n', file=sys.stderr)

        print('## System Information', file=sys.stderr)
        print_diagnostic_info()

        self.has_error = True
        self.pool.terminate()

    def passwords(self):
        self.lock.acquire()
        passwords = list(islice(self.iterable, self.chunk_size))
        self.lock.release()
        return passwords

    def crack(self, thread_id):
        """Attempt to crack a number of passwords"""
        time.sleep(0.1 * thread_id)

        attempts = 0
        password = ''
        passwords = self.passwords()
        thread_id = str(thread_id).rjust(len(str(self.thread_count)), '0')

        while any(passwords) and not self.password and not self.has_error:
            for password in passwords:
                if isinstance(password, bytes):
                    password = b2s(password).strip()

                attempts += 1

                with Popen([
                    'steghide',
                    'extract',
                    '-sf', self.file,
                    '-xf', self.output,
                    '-p', password,
                    '-f'
                ], stdout=PIPE, stderr=PIPE) as proc:
                    proc.wait()

                    if self.verbose:
                        self.lock.acquire()
                        sys.stderr.write(f'[Thread {thread_id}, password {password!r}]: ')
                        shutil.copyfileobj(b2s_file(proc.stderr), sys.stderr)
                        shutil.copyfileobj(b2s_file(proc.stdout), sys.stderr)
                        self.lock.release()

                if proc.returncode == 0:
                    self.password = password
                    self.pool.terminate()
                    break

            if not self.quiet:
                self.lock.acquire()

                self.attempts += attempts
                percentage = self.attempts * 100 / self.line_count

                print((
                    f"{self.attempts}/{self.line_count} ({percentage:.2f}%) "
                    f"Attempted: {password[0:20].ljust(20, ' ')}".strip()
                ), end='\r', flush=True, file=sys.stderr)

                self.lock.release()

            passwords = self.passwords()
            attempts = 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine
import re
import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

NAME = 'stegcracker'

DESCRIPTION = (
    'Steganography brute-force utility to '
    'uncover hidden data inside files.')

URL = 'https://github.com/Paradoxis/StegCracker'
EMAIL = 'luke@paradoxis.nl'
AUTHOR = 'Luke Paris (Paradoxis)'
REQUIRES_PYTHON = '>=3.6.0'

REQUIRED = []

EXTRAS_REQUIRE = {
    'test': [
        'coverage',
        'codecov'
    ],
    'dev': [
        'twine',
        'wheel',
        'setuptools>=40.6.3'
    ]
}

ENTRY_POINTS = {
    'console_scripts': [
        'stegcracker=stegcracker.__main__:main',
    ]
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = ('\n' + f.read()).strip()

with io.open(os.path.join(here, NAME.replace('-', '_'), '__init__.py'), encoding='utf-8') as init:
    VERSION = re.search(r'__version__ = [\'"]([\d.]+)[\'"]', init.read()).group(1)

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    with open(os.path.join(here, NAME, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds...')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution...')
        os.system('{0} setup.py sdist bdist_wheel'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine...')
        os.system('twine upload dist/*')

        self.status('Pushing git tags...')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    extras_require=EXTRAS_REQUIRE,
    entry_points=ENTRY_POINTS,
    install_requires=REQUIRED,
    include_package_data=True,
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    cmdclass={
        'upload': UploadCommand,
    }
)


import sys
from datetime import datetime
from argparse import ArgumentParser
from distutils.spawn import find_executable
from os.path import isfile

from stegcracker import __url__, __version__, __description__
from stegcracker.cracker import Cracker
from stegcracker.helpers import error, wc, handle_interrupt, DevNull, log, CustomHelpFormatter


DEFAULT_WORDLIST_PATH = '/usr/share/wordlists/rockyou.txt'


@handle_interrupt
def main():
    """Main entry point of the application"""

    if any(i in sys.argv for i in ('-q', '--quiet', '--stfu')):
        if any(i in sys.argv for i in ('-V', '--verbose')):
            return error(
                "Arguments 'verbose' and 'quiet', are mutually exclusive. "
                "You have to choose either one of them, not both.")

        sys.stderr = DevNull()

    if any(i in sys.argv for i in ('-v', '--version')):
        return print(__version__)

    log(f'StegCracker {__version__} - ({__url__})',)
    log(f'Copyright (c) {datetime.now().year} - Luke Paris (Paradoxis)')
    log('')
    log('StegCracker has been retired following the release of StegSeek, which ')
    log('will blast through the rockyou.txt wordlist within 1.9 second as opposed ')
    log('to StegCracker which takes ~5 hours.')
    log('')
    log('StegSeek can be found at: https://github.com/RickdeJager/stegseek')
    log('')

    args = ArgumentParser(
        usage='stegcracker <file> [<wordlist>]',
        formatter_class=CustomHelpFormatter,
        description=__description__)

    args.add_argument('file', action='store', help=(
        f'Input file you think contains hidden information and wish to crack. '
        f'Note: Stegcracker only accepts the following file types: '
        f'{", ".join(Cracker.SUPPORTED_FILES)}'))

    args.add_argument('wordlist', action='store', nargs='?', help=(
        'Wordlist containing the one or more passwords (one password per line). '
        'If no password list is supplied, this will default to the rockyou.txt '
        'wordlist on Kali Linux.'
    ), default=None)

    args.add_argument('-o', '--output', default=None, help=(
        'Output file location, this will be the file the data will be written '
        'to on a successful cracked password. If no output location is '
        'specified, the default location will be the same filename with ".out" '
        'appended to the name.'))

    args.add_argument('-t', '--threads', type=int, default=16, help=(
        'Number of concurrent threads used to crack passwords with, increasing '
        'this number might lead to better performance. Default: 16'))

    args.add_argument('-c', '--chunk-size', type=int, default=64, help=(
        'Number of passwords loaded into memory per thread cycle. After each '
        'password of the chunk has been depleted a status update will be '
        'printed to the console with the attempted password. Default: 64'))

    args.add_argument('-q', '--quiet', '--stfu', action='store_true', help=(
        'Runs the program in "quiet mode", meaning no status updates or other '
        'output besides the cracked password will be echoed to the terminal. '
        'By default, all logging / error messages are printed to stderr (making '
        'piping to other processes easier).'
    ), default=False)

    args.add_argument('-v', '--version', action='store_true', help=(
        'Print the current version number and exit.'))

    args.add_argument('-V', '--verbose', action='store_true', help=(
        'Runs the program in "verbose mode", this will print additional '
        'debugging information (include this output when submitting bug '
        'reports). Cannot be used in conjunction with the "--quiet" argument.'
    ), default=False)

    args = args.parse_args()

    output = args.output or args.file + '.out'
    extension = args.file.split('.')[::-1][0].lower()

    if not find_executable('steghide'):
        return error(
            'Steghide does not appear to be installed, or has not been added to '
            'your current PATH, please install it using: "apt-get install '
            'steghide -y" or by downloading it from the official code '
            'repository: http://steghide.sourceforge.net/')

    if args.wordlist is None:
        if isfile(DEFAULT_WORDLIST_PATH):
            log('No wordlist was specified, using default rockyou.txt wordlist.')
            args.wordlist = DEFAULT_WORDLIST_PATH

        elif isfile(DEFAULT_WORDLIST_PATH + '.gz'):
            return error(
                f'No wordlist was specified, but a gzipped variant of the '
                f'rockyou.txt wordlist was found on your system. If you wish '
                f'to use it, please decompress it first using the following command: '
                f'gzip -d {DEFAULT_WORDLIST_PATH}.gz')

        else:
            return error(
                'No wordlist was specified. '
                'For more help, please run: stegcracker --help')

    if isfile(output):
        return error(f'Output file {output!r} already exists!')

    if not isfile(args.file):
        return error(f'Input file {args.file!r} does not exist!')

    if not isfile(args.wordlist):
        return error(f'Wordlist {args.wordlist!r} does not exist!')

    if args.wordlist.endswith('.gz'):
        return error(
            f"It appears you're using a gzipped variant of a wordlist, instead "
            f"of the actual wordlist itself. You can decompress the gzipped "
            f"using the following command: gzip -d {args.wordlist}")

    if extension not in Cracker.SUPPORTED_FILES:
        return error(
            f'Unsupported file type {extension!r}! Supported '
            f'extensions: {", ".join(Cracker.SUPPORTED_FILES)}')

    if not args.quiet:
        log(f'Counting lines in wordlist..')
        line_count = wc(args.wordlist)
    else:
        line_count = None

    with open(args.wordlist, mode='rb') as wordlist:
        cracker = Cracker(
            file=args.file,
            output=output,
            line_count=line_count,
            chunk_size=args.chunk_size,
            quiet=args.quiet,
            verbose=args.verbose,
            threads=args.threads)

        if not args.quiet:
            log(
                f'Attacking file {args.file!r} '
                f'with wordlist {args.wordlist!r}..')

        cracker.run(wordlist)

    if cracker.password:
        log(f'Successfully cracked file with password: {cracker.password}')
        log(f'Tried {cracker.attempts} passwords')
        log(f'Your file has been written to: {output}')
        print(cracker.password)

    elif cracker.has_error:
        return error('Terminating due to previous exception..')

    else:
        return error('Failed to crack file, ran out of passwords.')


if __name__ == '__main__':
    exit(main() or 0)
